@@ -1,5 +1,7 @@
 # -*- coding: utf-8 -*-
 # (c) 2018 Andreas Motl <andreas@hiveeyes.org>
+from __future__ import print_function
+from builtins import object
 import attr
 import json
 import logging
@@ -11,6 +13,7 @@ from phenodata.util import haversine_distance
 
 logger = logging.getLogger(__name__)
 
+
 @attr.s
 class DwdPhenoData(object):
     """
@@ -25,9 +28,9 @@ class DwdPhenoData(object):
     The lists of phenological stations can be found in the list of phenology `annual reporters`_
     and the list of phenology `immediate reporters`_.
 
-    .. _cdc-readme: ftp://ftp-cdc.dwd.de/pub/CDC/Readme_intro_CDC_ftp.txt
-    .. _annual reporters: ftp://ftp-cdc.dwd.de/pub/CDC/help/PH_Beschreibung_Phaenologie_Stationen_Jahresmelder.txt
-    .. _immediate reporters: ftp://ftp-cdc.dwd.de/pub/CDC/help/PH_Beschreibung_Phaenologie_Stationen_Sofortmelder.txt
+    .. _cdc-readme: ftp://opendata.dwd.de/climate_environment/CDC/Readme_intro_CDC_ftp.txt
+    .. _annual reporters: ftp://opendata.dwd.de/climate_environment/CDC/help/PH_Beschreibung_Phaenologie_Stationen_Jahresmelder.txt
+    .. _immediate reporters: ftp://opendata.dwd.de/climate_environment/CDC/help/PH_Beschreibung_Phaenologie_Stationen_Sofortmelder.txt
 
     """
 
@@ -68,11 +71,11 @@ class DwdPhenoData(object):
     def get_quality_bytes(self):
         """
         Return DataFrame with quality bytes information
-        ftp://ftp-cdc.dwd.de/pub/CDC/observations_germany/climate/subdaily/standard_format/qualitaetsbytes.pdf
+        ftp://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/subdaily/standard_format/qualitaetsbytes.pdf
         """
         return self.cdc.get_dataframe(path='/help/PH_Beschreibung_Phaenologie_Qualitaetsbyte.txt', index_column=0)
 
-    def get_stations(self, all=False):
+    def get_stations(self, filter=None, all=False):
         """
         Return DataFrame with stations information.
         """
@@ -94,6 +97,27 @@ class DwdPhenoData(object):
         # Appropriately coerce geolocation values to float
         #dataframe_coerce_columns(data, ['geograph.Breite', 'geograph.Laenge'], float)
 
+        # A. Basic filtering
+
+        if filter:
+            # Build "boolean indexing" filter expression from multiple ID-based criteria
+            # https://pandas.pydata.org/pandas-docs/stable/indexing.html#boolean-indexing
+            reference_fields = [
+                'Stationsname',
+                'Naturraumgruppe',
+                'Naturraum',
+                'Bundesland',
+            ]
+
+            expression = False
+            for reference_field in reference_fields:
+                #print(reference_field, filter)
+                expression |= data[reference_field].str.contains(filter, case=False)
+
+            # Apply filter expression to DataFrame
+            if type(expression) is not bool:
+                data = data[expression]
+
         return data
 
     def nearest_station(self, latitude, longitude, all=False):
@@ -154,13 +178,24 @@ class DwdPhenoData(object):
         # Acquire data
         observations = self.query(partition=options['partition'], files=options['filename'])
 
+        # Sanity checks
+        if observations is None:
+            return
+
         # Filter data
         observations = self.flux(observations, criteria=options)
 
+        # Sanity checks
+        if observations is None:
+            logger.warning('No results found')
+            return
+
         # Optionally humanize DataFrame
         if humanize:
             megaframe = self.create_megaframe(observations)
             observations = self.humanizer.get_observations(megaframe)
+
+        # or pass-through with minor cosmetic amendments
         else:
             observations['Eintrittsdatum'] = observations['Eintrittsdatum'].astype(str)
 
@@ -179,6 +214,10 @@ class DwdPhenoData(object):
         # Get current observations
         observations = self.get_observations(options)
 
+        # Sanity checks
+        if observations is None:
+            return
+
         # Group by station, species and phase
         # https://pandas.pydata.org/pandas-docs/stable/groupby.html
         grouped = observations.groupby(['Stations_id', 'Objekt_id', 'Phase_id'])
@@ -194,25 +233,26 @@ class DwdPhenoData(object):
         real_dates = pd.to_datetime(datetime.today().year * 1000 + forecast['Jultag'], errors='coerce', format='%Y%j')
         forecast.insert(0, 'Datum', real_dates)
 
+        # Resolve index column to real columns
+        station_ids = []
+        species_ids = []
+        phase_ids = []
+        for index, row in forecast.iterrows():
+            # Decode index column and collect its components
+            station_id, species_id, phase_id = index
+            station_ids.append(station_id)
+            species_ids.append(species_id)
+            phase_ids.append(phase_id)
+
+        forecast['Stations_id'], forecast['Objekt_id'], forecast['Phase_id'] = station_ids, species_ids, phase_ids
+
+
         # Optionally humanize DataFrame
         if humanize:
-            station_ids = []
-            species_ids = []
-            phase_ids = []
-            for index, row in forecast.iterrows():
-
-                # Decode index column and collect its components
-                station_id, species_id, phase_id = index
-                station_ids.append(station_id)
-                species_ids.append(species_id)
-                phase_ids.append(phase_id)
-
-            # Re-add index components as real columns
-            forecast['Stations_id'], forecast['Objekt_id'], forecast['Phase_id'] = station_ids, species_ids, phase_ids
-
-            megaframe = self.create_megaframe(forecast)
+            megaframe = self.create_megaframe(forecast, drop_index_columns=True)
             forecast = self.humanizer.get_forecast(megaframe)
 
+        # or pass-through with minor cosmetic amendments
         else:
             forecast['Datum'] = forecast['Datum'].astype(str)
 
@@ -233,7 +273,6 @@ class DwdPhenoData(object):
         # Search FTP server
         paths = self.scan_files(partition, include=files, field='url')
 
-
         logger.info('Starting data acquisition with {} files'.format(len(paths)))
 
         # The main DataFrame object
@@ -242,23 +281,20 @@ class DwdPhenoData(object):
         # Load multiple files into single DataFrame
         for path in tqdm(paths, ncols=80):
 
-            # Skip invalid files
-            if 'Kulturpflanze_Ruebe_akt' in path:
-                logger.warning('Skipping file "{}" due to invalid header format (all caps)'.format(path))
-                continue
+            logger.debug('Processing file "{}"'.format(path))
 
             # Acquire DataFrame from CSV data
             data = self.cdc.get_dataframe(path, coerce_int=True)
 
             # Sanity checks
             if data is None:
+                logger.warning('File "{}" is empty')
                 continue
 
             # Coerce "Eintrittsdatum" column into date format
             data['Eintrittsdatum'] = pd.to_datetime(data['Eintrittsdatum'], errors='coerce', format='%Y%m%d')
 
-            # Append to DataFrame
-            results = results.append(data)
+            results = results.append(data, sort=False)
 
         # Sanity checks
         if results.empty:
@@ -270,25 +306,30 @@ class DwdPhenoData(object):
 
         return results
 
-    def create_megaframe(self, frame):
+    def create_megaframe(self, frame, drop_index_columns=False):
 
         # https://pandas.pydata.org/pandas-docs/stable/merging.html#database-style-dataframe-joining-merging
 
+        # Prevent errors like::
+        #   ValueError: 'Stations_id' is both an index level and a column label, which is ambiguous.
+        if drop_index_columns:
+            frame.drop(columns=['Stations_id'], axis='columns', inplace=True)
+
         # Stations
-        frame = pd.merge(frame, self.get_stations(), left_on='Stations_id', right_index=True)
+        frame = pd.merge(frame, self.get_stations(), left_on='Stations_id', right_on='Stations_id')
 
         # Species
-        frame = pd.merge(frame, self.get_species(), left_on='Objekt_id', right_index=True)
+        frame = pd.merge(frame, self.get_species(), left_on='Objekt_id', right_on='Objekt_ID')
 
         # Phases
-        frame = pd.merge(frame, self.get_phases(), left_on='Phase_id', right_index=True)
+        frame = pd.merge(frame, self.get_phases(), left_on='Phase_id', right_on='Phase_ID')
 
         # Quality level
         if 'Qualitaetsniveau' in frame:
             frame = pd.merge(frame, self.get_quality_levels(), left_on='Qualitaetsniveau', right_index=True)
 
         # Quality byte
-        # ftp://ftp-cdc.dwd.de/pub/CDC/observations_germany/climate/subdaily/standard_format/qualitaetsbytes.pdf
+        # ftp://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/subdaily/standard_format/qualitaetsbytes.pdf
         if 'Eintrittsdatum_QB' in frame:
             frame = pd.merge(frame, self.get_quality_bytes(), left_on='Eintrittsdatum_QB', right_index=True)
 
@@ -320,18 +361,21 @@ class DwdPhenoData(object):
         # Build "boolean indexing" filter expression from multiple ID-based criteria
         # https://pandas.pydata.org/pandas-docs/stable/indexing.html#boolean-indexing
         isin_map = {
-            'year': results.Referenzjahr,
-            'quality-level': results.Qualitaetsniveau,
-            'quality-byte': results.Eintrittsdatum_QB,
-            'station-id': results.Stations_id,
-            'species-id': results.Objekt_id,
-            'phase-id': results.Phase_id,
+            'year': 'Referenzjahr',
+            'quality-level': 'Qualitaetsniveau',
+            'quality-byte': 'Eintrittsdatum_QB',
+            'station-id': 'Stations_id',
+            'species-id': 'Objekt_id',
+            'phase-id': 'Phase_id',
         }
 
         # Lowlevel filtering based on IDs
         # For each designated field, add ``.isin`` criteria to "boolean index" expression
         expression = True
-        for key, reference in isin_map.items():
+        for key, field in list(isin_map.items()):
+            if field not in results:
+                continue
+            reference = results[field]
             if criteria[key]:
                 expression &= reference.isin(criteria[key])
 
@@ -357,7 +401,7 @@ class DwdPhenoData(object):
 
         expression = True
         is_megaframe = False
-        for field, reference_fields in patterns_map.items():
+        for field, reference_fields in list(patterns_map.items()):
 
             if criteria[field]:
 
@@ -397,7 +441,7 @@ class DwdPhenoData(object):
             url, subdir=partition,
             include=include,
             include_base=['PH_(Sofort|Jahres)melder.+\.txt'],
-            exclude_base=['PH_Beschreibung', 'Spezifizierung'],
+            exclude_base=['PH_Beschreibung', 'Spezifizierung', 'PH_.+_Notiz'],
         )
 
         # Return entries if projection to field not requested
@@ -453,6 +497,7 @@ class DwdPhenoDataHumanizer(object):
         # Build fresh DataFrame with designated order of columns
         canvas['Jahr'] = frame['Referenzjahr']
         canvas['Datum'] = frame['Eintrittsdatum'].astype(str)
+        canvas['Tag'] = frame['Jultag']
         canvas['Spezies'] = species
         canvas['Phase'] = phases
         canvas['Station'] = stations
@@ -469,6 +514,7 @@ class DwdPhenoDataHumanizer(object):
 
         # Build fresh DataFrame with designated order of columns
         canvas['Datum'] = frame['Datum'].astype(str)
+        canvas['Tag'] = frame['Jultag']
         canvas['Spezies'] = species
         canvas['Phase'] = phases
         canvas['Station'] = stations
@@ -541,7 +587,7 @@ class DwdPhenoDataHumanizer(object):
                 quality_levels.append(ql_label)
 
             # Qualitaetsbyte
-            # ftp://ftp-cdc.dwd.de/pub/CDC/observations_germany/climate/subdaily/standard_format/qualitaetsbytes.pdf
+            # ftp://opendata.dwd.de/climate_environment/CDC/observations_germany/climate/subdaily/standard_format/qualitaetsbytes.pdf
             if 'Eintrittsdatum_QB' in row:
                 qb_label = row.get('Beschreibung_y', '')
                 if self.show_ids:
